# LigandEfficiency.py - Ligand efficiency plugin script for StarDrop
#
# Copyright (C) Optibrium Ltd 2023
# stardrop-support@optibrium.com
#
# This script allows the user to calculate a prediction of Ligand
# Efficiency for a given column of target concentration values
#
# For advice or feedback, contact stardrop-support@optibrium.com

import os
import re
import textwrap
from math import log10
from typing import Any, Optional, Callable

import app
import stardropunits

import stardrop_helper_functions.logfile_handling as logfile
import stardrop_helper_functions.dataset_import_export as dimex
from stardrop_helper_functions.qt_widgets import DisplayGUIs, PromptGUIs


"""
The mathematical equation used within this script is -1.4*log[pKi]/N, where

            - pKi = log_{10}[Ki], and Ki = compound concentration

            - N = number of heavy (i.e., non-hydrogen) atoms
"""


VERSION_NUMBER_STR = '2.0'


def get_relevant_column_names(stardrop_dataset: dict) -> tuple[Optional[str], Optional[str]]:
    """
    This function extracts the relevant column names containing the Structure strings and
    concentrations needed to calculate ligand efficiencies from the given data set.

    @param: The data set containing the Structure string and concentrations to calculate
            ligand efficiencies over
    @returns: The following column names (which may be None-valued if this function can't find
              any appropriate column names):
                1) smiles_column_name: The name of the column containing Structure strings
                2) concentration_column_name: The name of the column containing compound concentrations
                                         (this function will try to use pKI or µM columns if it can find
                                         them, but will default to generical numerical columns if it can't
                                         otherwise find any)
    """

    def select_column(column_names: list[str],
                      column_type_description: str,
                      always_select: bool) -> Optional[str]:
        """
        An internally defined local function which is meant to make the code within
        get_relevant_column_names() easier to understand.

        @param: column_names: The list of column names which the user must select from
        @param: column_type_description: The description of the column type to be selected
        @param: always_select: Whether or not to make the user always select a column, even if there's
                               only one column
        @returns: A user-selected column
        """

        selected_column_name = None

        if column_names:

            custom_explanation_msg = 'Select column with {column_type} units' if always_select else 'Multiple {column_type} columns were found'  # noqa: E501

            selected_column_name = \
                PromptGUIs.select_column(column_names,
                                         f'Choose {column_type_description} column',
                                         textwrap.fill(custom_explanation_msg.format(column_type=column_type_description),        # noqa: E501
                                                       width=50)) if len(column_names) > 1 or always_select else column_names[0]  # noqa: E501

            if not selected_column_name:
                logfile.record_error(f'User cancelled out of selecting a {column_type_description} column')

        else:
            err_msg = f'Could not find a {column_type_description} column'
            logfile.record_error(err_msg)
            DisplayGUIs.warning_window(err_msg, 'Could not find column')

        return selected_column_name

    smiles_column_name = \
        select_column([column_name for column_name, column_type in stardrop_dataset['column type'].items() if column_type == app.mol],  # noqa: E501
                      'Structure',
                      False)

    concentration_column_name = \
        select_column([column_name for column_name, column_type in stardrop_dataset['column type'].items() if column_type == app.cont],  # noqa: E501
                      'pKi or µM Ki',
                      True) if smiles_column_name else None

    return [smiles_column_name,
            concentration_column_name]


def calculate_ligand_efficiencies(appui, stardrop_dataset: dict[str, Any]) -> None:
    """
    This function calculates the ligand efficiencies using the extracted Structure strings
    and concentrations, and then exports the calculations out to the StarDrop GUI.

    @param: appui: The StarDrop GUI
    @param: The data set containing the Structure strings and concentrations to calculate
            ligand efficiencies over
    """

    [smiles_column_name,
     concentration_column_name] = \
        get_relevant_column_names(stardrop_dataset)

    if smiles_column_name and concentration_column_name:

        smiles_column        = stardrop_dataset['column'       ][       smiles_column_name]           # noqa: E201, E202, E221, E501
        concentration_column = stardrop_dataset['column'       ][concentration_column_name]           # noqa: E202
        concentration_units  = stardrop_dataset['column format'][concentration_column_name]['units']  # noqa: E221

        heavy_atom_pattern = re.compile('[CcOoNnSsBPFI]|\[\w+\]')  # noqa: W605

        calculate_LE_from_pKi   = lambda   pKi, heavy_atom_count: 1.4*pKi/heavy_atom_count                  # noqa: E221, E271, E501, E731
        calculate_LE_from_uM_Ki = lambda uM_Ki, heavy_atom_count: 1.4*(-log10(uM_Ki) + 6)/heavy_atom_count  # noqa: E731

        def calculate_ligand_efficiency_tuple(Ki_tuple: tuple[float, float],
                                              heavy_atom_count: int,
                                              concentration_units: int,
                                              calculate_LE_from_pKi: Callable[[float, int], float],
                                              calculate_LE_from_uM_Ki: Callable[[float, int], float]) -> tuple[float, float, str]:  # noqa: E501

            if concentration_units == stardropunits.MICRO:
                LE_tuple = (calculate_LE_from_uM_Ki(Ki_tuple[0], heavy_atom_count), 0, '')
            else:
                LE_tuple = \
                    (calculate_LE_from_pKi(Ki_tuple[0], heavy_atom_count),
                     calculate_LE_from_pKi(Ki_tuple[1], heavy_atom_count),
                     '')

            return LE_tuple

        ligand_efficiency_column = \
            [calculate_ligand_efficiency_tuple(concentration_tuple,
                                               len(heavy_atom_pattern.findall(smiles)),
                                               concentration_units,
                                               calculate_LE_from_pKi,
                                               calculate_LE_from_uM_Ki) if smiles and concentration_tuple else None for smiles, concentration_tuple in zip(smiles_column,          # noqa: E501
                                                                                                                                                           concentration_column)]  # noqa: E501

        ligand_efficiency_column_name = f"Ligand Efficiency ({concentration_column_name})"
        stardrop_dataset['column'][ligand_efficiency_column_name] = ligand_efficiency_column
        stardrop_dataset['column type'][ligand_efficiency_column_name] = app.cont
        stardrop_dataset['column index'][ligand_efficiency_column_name] = stardrop_dataset['column index'][concentration_column_name] + 1  # noqa: E501
        stardrop_dataset['column format'][ligand_efficiency_column_name] = None

        dimex.export_stardrop_column(appui, stardrop_dataset, ligand_efficiency_column_name)


def ligand_efficiency(appui) -> None:
    """
    This function is the main function which is called from the StarDrop GUI.

    @param: appui: The StarDrop GUI
    """

    logfile.initialize_logfile('Ligand Efficiency',
                               VERSION_NUMBER_STR,
                               os.getcwd(),
                               'lig_eff')

    try:

        stardrop_dataset = \
            dimex.import_stardrop_dataset(appui,
                                          preferred_mol_str_type='SMILES')

        if stardrop_dataset:
            calculate_ligand_efficiencies(appui, stardrop_dataset)

    except Exception as err:
        logfile.record_and_raise_exception(err)

    logfile.finish_logfile()


def get_stardrop_definitions():

    definitions = \
        [{'script_name': 'Calculate Ligand Efficiency',
          'callback': ligand_efficiency}]

    return definitions
